//------------------------------ $Keywords ----------------------------------
// GPGee - GNU Privacy Guard Explorer Extension
// GPGeeUtility.cpp - Miscelaneous utility functions
// Copyright 2005, Kurt Fitzner <kfitzner@excelcia.org>
//---------------------------------------------------------------------------
// This file is part of GPGee.
//
// GPGee is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GPGee is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// VCS: $Version: 1 $ $Revision: 1 $
/*
$History: **** V 1.0 by kfitzner ****
$History: * gpgeeutility.cpp - 2005-05-13 11:14:16 PM - 10303 Bytes
$History: * gpgeeutility.h - 2005-05-12 9:05:36 AM - 5764 Bytes
$History: * Initial check-in
$History: **** Latest ** V 1.1 by kfitzner ** 2005-05-17 2:16:33 AM ****
$History: * Add persistent key cache
*/
//----------------------------  $NoKeywords ---------------------------------


//---------------------------------------------------------------------------
// File Notes:
//---------------------------------------------------------------------------
// 21 March 2005 - Kurt Fitzner <kfitzner@excelcia.org>
//
// This unit is simply a respository for utility functions that are used by
// more than one form or unit.
//---------------------------------------------------------------------------
// 6 May 2006 - Kurt Fitzner (kfitzner@excelcia.org>
//
// I've used this unit as the location for the language file message loading
// and formatting functions GetMessage() and FormMessage().  The former
// loads a string from the current language DLL, the latter loads and formats
// the string through vprintf.
//---------------------------------------------------------------------------
#include <vcl.h>
#include <FileCtrl.hpp>
#pragma hdrstop

#include "gpgme/gpgme.h"
#include "GPGeeExceptions.h"
#include "TProgramLog.h"
#include "TConfiguration.h"
#include "TProgressForm.h"
#include "GPGeeUtility.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
extern TConfiguration *GPGeeConfig;
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// Take a filename and reduce it so that it fits inside the space given on
// an edit box.
//
AnsiString __fastcall MinimizeName(AnsiString FileName, TEdit *Edit)
{
  TCanvas *FileNameCanvas;

  __ENTERFUNCTION__;
  LOG1(LOG_DEBUG, "Minimizing filename \"%s\"", FileName.c_str());
  FileNameCanvas = new TCanvas;
  FileNameCanvas->Handle = GetDC(GetActiveWindow());
  FileNameCanvas->Font->Assign(Edit->Font);
  Edit->Text = Filectrl::MinimizeName(FileName, FileNameCanvas, Edit->Width-4);
  ReleaseDC(GetActiveWindow(),FileNameCanvas->Handle);
  __RETURNFUNCTION(Edit->Text);
}  // AnsiString __fastcall MinimizeName(AnsiString FileName, TControl *Control)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Take a filename and reduce it so that it fits inside the space given on
// an edit box.
//
AnsiString __fastcall MinimizeName(AnsiString FileName, TFont *Font, int Width)
{
  AnsiString retval;
  TCanvas *FileNameCanvas;

  __ENTERFUNCTION__;
  LOG1(LOG_DEBUG, "Minimizing filename \"%s\"", FileName.c_str());
  FileNameCanvas = new TCanvas;
  FileNameCanvas->Handle = GetDC(GetActiveWindow());
  FileNameCanvas->Font->Assign(Font);
  retval = Filectrl::MinimizeName(FileName, FileNameCanvas, Width);
  ReleaseDC(GetActiveWindow(),FileNameCanvas->Handle);
  __RETURNFUNCTION(retval);
}  // AnsiString __fastcall MinimizeName(AnsiString FileName, TControl *Control)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Identify a GnuPG file.  This is done in one of two ways.  If the file is
// ascii armored, it will have a .asc extension.  These files have text
// headers which identify their type.  If the file has a .sig extension,
// then it is a detached signature.  If it has a .gpg extension, then it
// could be almost anything.  These are treated simply as "messages" which
// a decrypt operation will be attempted on.
//
gpg_file_type __fastcall IdentifyFile(AnsiString FileName)
{
  AnsiString Extension;
  gpg_file_type retval = GPG_UNKNOWN;

  __ENTERFUNCTION__;
  LOG1(LOG_DEBUG, "Identifying file \"%s\"", FileName.c_str());
  Extension = ExtractFileExt(FileName);
  if (Extension.AnsiCompareIC(".gpg") == 0)
    retval = GPG_MESSAGE;
  else if (Extension.AnsiCompareIC(".sig") == 0)
    retval = GPG_SIGNATURE;
  else if (Extension.AnsiCompareIC(".asc") == 0) {
    gpgme_pgptype_t pgptype;
    if (gpgme_file_get_pgptype(FileName.c_str(), &pgptype) == GPGME_No_Error)
      pgptype &= (GPGME_TYPE_MESSAGE | GPGME_TYPE_SIG | GPGME_TYPE_CLEARSIG);
      switch (pgptype) {
        case GPGME_TYPE_MESSAGE:
          retval = GPG_MESSAGE;
          break;
        case GPGME_TYPE_SIG:
          retval = GPG_SIGNATURE;
          break;
        case GPGME_TYPE_CLEARSIG:
        case GPGME_TYPE_CLEARSIG | GPGME_TYPE_SIG:
          retval = GPG_CLEARSIGNED;
          break;
        default:
          LOG2(LOG_WARNING, "Could not identify .asc file \"%s\": pgptype=0x%X.", FileName.c_str(), pgptype);
          retval = GPG_UNKNOWN;
      }  // switch (pgptype)
  }  // else if (Extension.AnsiCompareIC(".asc") == 0)
  __RETURNFUNCTION(retval);
}  // int __fastcall IdentifyFile(AnsiString FileName)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Convert GnuPG dates to TDateTime
//
TDateTime __fastcall GPGToDateTime(unsigned long GPGDate, bool LocalTime) {
  TIME_ZONE_INFORMATION tzInfo;
  DWORD tzResult;
  static double tzBias = -1.0;
  TDateTime retval;

  __ENTERFUNCTION__;

  // GnuPG returns timestamps in GMT - we convert that to local time here, so we need to get the time zone info
  if (tzBias == -1.0) {
    tzResult = GetTimeZoneInformation(&tzInfo);
    if (tzResult == TIME_ZONE_ID_STANDARD)
      tzBias = ((double)tzInfo.Bias + (double)tzInfo.StandardBias) / 1440.0;
    else if (tzResult == TIME_ZONE_ID_DAYLIGHT)
      tzBias = ((double)tzInfo.Bias + (double)tzInfo.DaylightBias) / 1440.0;
    else
      tzBias = 0.0;
  }  // if (tzBias = -1.0)
  retval = (double)GPGDate / (24.0 * 60.0 * 60.0) + UnixDateDelta - (LocalTime?tzBias:0.0);
  LOG2(LOG_DEBUG, "GPG date %d converted to %s.", GPGDate, retval.DateTimeString().c_str());

  __RETURNFUNCTION(retval);
}  // TDateTime __fastcall GPGToDateTime(unsigned long GPGDate)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Return a short string description for a GnuPG public key id number
//
AnsiString __fastcall KeyTypeToString(gpgme_pk_cipher_t KeyType)
{
  AnsiString retval;

  __ENTERFUNCTION__;

  switch (KeyType) {
    case GPGME_PK_RSA:
      retval = "RSA"; break;
    case GPGME_PK_RSA_E:
      retval = "RSAe"; break;
    case GPGME_PK_RSA_S:
      retval = "RSAs"; break;
    case GPGME_PK_ELG_E:
      retval = "ELG"; break;
    case GPGME_PK_DSA:
      retval = "DSA"; break;
    case GPGME_PK_ELG_ES:
      retval = "ELGse"; break;
  }  // switch (KeyType)

  __RETURNFUNCTION(retval);
}  // AnsiString __fastcall KeyTypeToString(gpgme_pk_cipher_t KeyType)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
AnsiString __fastcall TrustToString(gpgme_validity_t Trust)
{
  AnsiString retval;
  __ENTERFUNCTION__;
  retval = GetMessage(MSG_GPGME_TRUST_UNKNOWN + Trust);
  __RETURNFUNCTION(retval);
}  // AnsiString __fastcall TrustToString(GpgmeOwnertrust Trust)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
AnsiString __fastcall DigestToString(gpgme_md_t Digest)
{
  AnsiString retval;
  __ENTERFUNCTION__;
  retval = GetMessage(MSG_GPGME_MD_UNKNOWN + Digest);
  __RETURNFUNCTION(retval);
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Return a string from the language resource module
//
AnsiString __fastcall GetMessage(int ident)
{
  static HINSTANCE ResourceHInstance = 0;
  AnsiString Message;

  __ENTERFUNCTION__;

  LOG1(LOG_DEBUG, "Retrieving message #%d from the language string table.", ident);
  if (!ResourceHInstance) {
    ResourceHInstance = (HINSTANCE)FindResourceHInstance((unsigned int)HInstance);
    if (!ResourceHInstance)
      ResourceHInstance = HInstance;
  }  // if (!ResourceHInstance)
  Message.LoadString(ResourceHInstance, ident);

  __RETURNFUNCTION(Message);
}  // AnsiString GetMessage(int ident)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Return a formatted message composed of a language resource string that
// has been run through vprintf()
//
AnsiString FormMessage(int ident, ...)
{
  va_list valist;
  AnsiString Message;

  __ENTERFUNCTION__;
  va_start(valist, ident);
  Message.vprintf(GetMessage(ident).c_str(),valist);
  va_end(valist);
  __RETURNFUNCTION(Message);
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void ProgressCallback(void *opaque, const char *what, int type, unsigned current, unsigned total)
{
  if (!total)
    return;
  if (!ProgressForm)
    ProgressForm = new TProgressForm(NULL, "GPGee is caching your " + String((char *)opaque) + " keyring...");
  ProgressForm->SetPosition(current, total);
  ProgressForm->Show();
  Application->ProcessMessages();
}  // void ProgressCallback(void *opaque, const char *what, int type, unsigned current, unsigned total)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Return a GPGME key cache.  This function will keep a persistent cache
// if GPGee is configured to do so, otherwise it will return a newly
// initialized cache each time it is called.
//
gpgme_keycache_t __fastcall GetKeyCache(bool CacheSecretKeys, bool Cleanup) {
  static gpgme_keycache_t cachePublic = NULL;
  static gpgme_keycache_t cacheSecret = NULL;
  static int PublicCacheStamp = -1;
  static int SecretCacheStamp = -1;
  gpgme_keycache_t *pCache;
  int *pStamp;
  int CurrentStamp;
  AnsiString CacheFile;
  gpgme_error_t err;
  bool PersistentCache;
  AnsiString CountConfig;

  __ENTERFUNCTION__;

  // Cleanup the old caches if they exist and we are called to do so.
  if (Cleanup) {
    if (cachePublic) {
      gpgme_keycache_release(cachePublic);
      cachePublic = NULL;
    }  // if (cachePublic);
    if (cacheSecret) {
      gpgme_keycache_release(cacheSecret);
      cacheSecret = NULL;
    }  // if (cacheSecret);
    __RETURNFUNCTION(NULL);
  }  // if (Cleanup)

  try {
    PersistentCache = GPGeeConfig->Values["Cache Keys"];
    CountConfig = CacheSecretKeys?"Secring Count":"Pubring Count";

    // Configure our cache and timestamp pointers and get the keyring filename.
    if (CacheSecretKeys) {
      pCache = &cacheSecret;
      pStamp = &SecretCacheStamp;
      CacheFile = GPGeeConfig->Values["Secret Keyring"];
    } else {
      pCache = &cachePublic;
      pStamp = &PublicCacheStamp;
      CacheFile = GPGeeConfig->Values["Keyring"];
    }  // if (CacheSecretKeys);

    if (PersistentCache && !CacheFile.IsEmpty()) {
      CurrentStamp = FileAge(CacheFile);
      if (CurrentStamp == *pStamp && *pCache) {
        LOG1(LOG_DEBUG, "Keyring timestamp matches and existing cache exists - reusing %s key cache.", CacheSecretKeys?"secret":"public");
        __RETURNFUNCTION(*pCache);
      } else {
        // If we get here it means we are configured to use a persistent cache but either the cache
        // hasn't been initialized yet or the timestamps don't match so we store the current timestamp.
        if (CurrentStamp)
          *pStamp = CurrentStamp;
      }  // if (CurrentStamp == *pStamp && *pCache)
    }  // if (GPGeeConfig->Values["Cache Keys"])

    if (*pCache) {
      gpgme_keycache_release(*pCache);
      *pCache = NULL;
    }  // if (*pCache)
    err = gpgme_keycache_new(pCache);
    if (err)
      throw EGPGMEError(FormMessage(CacheSecretKeys?MSG_ERROR_GPGME_KEYCACHE_INIT_PUB:MSG_ERROR_GPGME_KEYCACHE_INIT_PUB,gpgme_strerror(err)));
    gpgme_keycache_set_cb(*pCache, ProgressCallback, CacheSecretKeys?"secret":"public", GPGeeConfig->Values[CountConfig]);
    err = gpgme_keycache_init(*pCache, NULL, CacheSecretKeys);
    GPGeeConfig->Values[CountConfig] = gpgme_keycache_count(*pCache);
    gpgme_keycache_set_cb(*pCache, NULL, NULL, 0);
    if (ProgressForm) {
      ProgressForm->Close();
      ProgressForm->Release();
      ProgressForm = NULL;
    }  // if (ProgressForm)
    if (err)
      throw EGPGMEError(FormMessage(CacheSecretKeys?MSG_ERROR_GPGME_KEYCACHE_LOAD_SEC:MSG_ERROR_GPGME_KEYCACHE_LOAD_PUB,gpgme_strerror(err)));    
    __RETURNFUNCTION(*pCache);
  }  // try

  catch (...) {
    if (cachePublic) {
      gpgme_keycache_release(cachePublic);
      cachePublic = NULL;
    }  // if (cachePublic);
    if (cacheSecret) {
      gpgme_keycache_release(cacheSecret);
      cacheSecret = NULL;
    }  // if (cacheSecret);
    __RETURNFUNCTION(NULL);
  }  // catch (...)

}  // gpgme_keycache_t __fastcall GetKeyCache(bool GetSecretKeys)
//---------------------------------------------------------------------------

